#include "Diffuse.hpp"
#include <fstream>

void Diffuse(weight_address_Data_T Resampled_Addresses[NUM_PARTICLES],
		Mem_SAMPLE Samples_in[NUM_PARTICLES],
		Mem_SAMPLE Samples_out[NUM_PARTICLES],
		Sample_abs_indx_Data_T Sample_abs_indx[NUM_PARTICLES],
		Sample_abs_indx_Data_T Sample_abs_indx_CPY[NUM_PARTICLES],
		Mem_Heuristic Heuristics[6300], BBOX all_box[6300],
		std_dev_Data_T trans_var, std_dev_Data_T rot_var) {
#pragma HLS DATAFLOW
	static hls::stream<u4_Data_T> diffuse_type_stream("diffuse_type_stream");
#pragma HLS STREAM variable=diffuse_type_stream depth=2 dim=1
	static hls::stream<Gauss_Data_T> Trans_Rand("trans_rand_stream");
#pragma HLS STREAM variable=Trans_Rand depth=2 dim=1
	static hls::stream<Gauss_Data_T> Rot_Rand("rot_rand_stream");
#pragma HLS STREAM variable=Rot_Rand depth=2 dim=1

	diffuse_type_generator(diffuse_type_stream);
	Box_muller(rot_var, trans_var, Trans_Rand, Rot_Rand);
	diffuse_case(diffuse_type_stream, Resampled_Addresses, Samples_in,
			Samples_out, Sample_abs_indx, Sample_abs_indx_CPY, Heuristics,
			all_box, Trans_Rand, Rot_Rand);

}

void diffuse_type_generator(hls::stream<u4_Data_T> & diffuse_type_stream) {

	static u15_Data_T lfsr = 347;

	//these one are needed otherwise they will be sinth as 32 or 64 bit constant !
	for (u10_Data_T i = 0; i < Num_Particles - 1; i++) {
#pragma HLS PIPELINE II=1
		u1_Data_T b_15 = lfsr.get_bit(15 - 15);
		u1_Data_T b_14 = lfsr.get_bit(15 - 14);
		u1_Data_T new_bit = b_15 ^ b_14;
		lfsr = lfsr >> 1;
		lfsr.set_bit(14, new_bit);
		ap_ufixed<15, 0> rand_01;
		rand_01.range() = lfsr; // u 1 - > 0

		u4_Data_T fill = 13 * rand_01;

		diffuse_type_stream.write(fill);
	}
#ifndef NDEBUG
	std::cout << "EXIT_DIFFUSE_TYPE_GENERATOR" << std::endl;
#endif
}

#ifdef NDEBUG
std::string fileFS = "Diffuse.txt";
std::ofstream FS_PROB(fileFS.c_str());
#endif

void diffuse_case(hls::stream<u4_Data_T> & diffuse_type_stream,
		weight_address_Data_T Resampled_Addresses[NUM_PARTICLES],
		Mem_SAMPLE Samples_in[NUM_PARTICLES],
		Mem_SAMPLE Samples_out[NUM_PARTICLES],
		Sample_abs_indx_Data_T Sample_abs_indx[NUM_PARTICLES],
		Sample_abs_indx_Data_T Sample_abs_indx_OUT[NUM_PARTICLES],
		Mem_Heuristic Heuristics[6300], BBOX all_box[6300],
		hls::stream<Gauss_Data_T> &Trans_Rand,
		hls::stream<Gauss_Data_T> &Rot_Rand) {

	int count_trans = 0;
	int count_rot = 0;
	for (weight_address_Data_T counter = 0; counter < NUM_PARTICLES - 1;
			counter++) {
		u4_Data_T type = diffuse_type_stream.read();
		weight_address_Data_T relative_addr = Resampled_Addresses[counter];
		// the new change !!
		Sample_abs_indx_Data_T abs_addr = Sample_abs_indx[relative_addr];
		Sample_abs_indx_Data_T new_abs_addr;
		int offset = 0;
		u2_Data_T direction = 0;
		ap_uint<8> h_ind;
		h_ind = all_box[abs_addr].h_index_w_index.range(16, 9);
		ap_int<9> w_ind = all_box[abs_addr].h_index_w_index.range(7, 0);
		ap_uint<8> map_width;
		ap_uint<8> map_height;
		ap_int<9> temp_ind;
		ap_fixed<14 + 5, 6> temp_angle;
		angle_Data_T TEMP_ADD;
		u2_Data_T rot_type;
#ifdef NDEBUG
//		printf("Diffuse particle num: %d with type: %d \n", counter, type);
		FS_PROB <<"Diffuse particle num: " << counter << " with type: "<<type;
#endif
		Sample_abs_indx_OUT[counter] = abs_addr;
		switch (type) {
		case 0:
			//translation on x
			Samples_out[counter].y_S = Samples_in[relative_addr].y_S;
			Samples_out[counter].z_S = Samples_in[relative_addr].z_S;
			Samples_out[counter].xroll_S = Samples_in[relative_addr].xroll_S;
			Samples_out[counter].ypitch_S = Samples_in[relative_addr].ypitch_S;
			Samples_out[counter].zyaw_S = Samples_in[relative_addr].zyaw_S;
#ifdef NDEBUG
			FS_PROB<<" on x direction";
#endif
			diffuse_translation(Samples_in[relative_addr].x_S,
					Samples_out[counter].x_S, Heuristics[abs_addr].x_min,
					Heuristics[abs_addr].x_max, Trans_Rand);
			count_trans++;
			break;
		case 1:
			//translation on y
			Samples_out[counter].x_S = Samples_in[relative_addr].x_S;
			Samples_out[counter].z_S = Samples_in[relative_addr].z_S;
			Samples_out[counter].xroll_S = Samples_in[relative_addr].xroll_S;
			Samples_out[counter].ypitch_S = Samples_in[relative_addr].ypitch_S;
			Samples_out[counter].zyaw_S = Samples_in[relative_addr].zyaw_S;
#ifdef NDEBUG
			FS_PROB<<" on y direction";
#endif
			diffuse_translation(Samples_in[relative_addr].y_S,
					Samples_out[counter].y_S, Heuristics[abs_addr].y_min,
					Heuristics[abs_addr].y_max, Trans_Rand);
			count_trans++;
			break;
		case 2:
#ifdef NDEBUG
			FS_PROB<<" on z direction";
#endif
			//translation on z
			Samples_out[counter].x_S = Samples_in[relative_addr].x_S;
			Samples_out[counter].y_S = Samples_in[relative_addr].y_S;
			Samples_out[counter].xroll_S = Samples_in[relative_addr].xroll_S;
			Samples_out[counter].ypitch_S = Samples_in[relative_addr].ypitch_S;
			Samples_out[counter].zyaw_S = Samples_in[relative_addr].zyaw_S;

			diffuse_translation(Samples_in[relative_addr].z_S,
					Samples_out[counter].z_S, Heuristics[abs_addr].z_min,
					Heuristics[abs_addr].z_max, Trans_Rand);
			count_trans++;
			break;
		case 3:
			//rotation on rot_X
#ifdef NDEBUG
			FS_PROB <<" on x_rot direction";
#endif
			Samples_out[counter].x_S = Samples_in[relative_addr].x_S;
			Samples_out[counter].y_S = Samples_in[relative_addr].y_S;
			Samples_out[counter].z_S = Samples_in[relative_addr].z_S;
			Samples_out[counter].ypitch_S = Samples_in[relative_addr].ypitch_S;
			Samples_out[counter].zyaw_S = Samples_in[relative_addr].zyaw_S;

			diffuse_rotation(Samples_in[relative_addr].xroll_S,
					Samples_out[counter].xroll_S, Rot_Rand);
#ifdef NDEBUG
			FS_PROB << " with workspace_min: " << Heuristics[abs_addr].x_min
			<< " workspace_max: " << Heuristics[abs_addr].x_max
			<< std::endl;
#endif
			count_rot++;
			break;
		case 4:
#ifdef NDEBUG
			FS_PROB <<" on y_rot direction";
#endif
			//rotation on rot_Y
			Samples_out[counter].x_S = Samples_in[relative_addr].x_S;
			Samples_out[counter].y_S = Samples_in[relative_addr].y_S;
			Samples_out[counter].z_S = Samples_in[relative_addr].z_S;
			Samples_out[counter].xroll_S = Samples_in[relative_addr].xroll_S;
			Samples_out[counter].zyaw_S = Samples_in[relative_addr].zyaw_S;

			diffuse_rotation(Samples_in[relative_addr].ypitch_S,
					Samples_out[counter].ypitch_S, Rot_Rand);
#ifdef NDEBUG
			FS_PROB << " with workspace_min: " << Heuristics[abs_addr].x_min
			<< " workspace_max: " << Heuristics[abs_addr].x_max
			<< std::endl;
#endif
			count_rot++;
			break;
		case 5:
			//rotation on rot_Z
#ifdef NDEBUG
			FS_PROB<<" on z_rot direction";
#endif
			Samples_out[counter].x_S = Samples_in[relative_addr].x_S;
			Samples_out[counter].y_S = Samples_in[relative_addr].y_S;
			Samples_out[counter].z_S = Samples_in[relative_addr].z_S;
			Samples_out[counter].xroll_S = Samples_in[relative_addr].xroll_S;
			Samples_out[counter].ypitch_S = Samples_in[relative_addr].ypitch_S;
			diffuse_rotation(Samples_in[relative_addr].zyaw_S,
					Samples_out[counter].zyaw_S, Rot_Rand);
#ifdef NDEBUG
			FS_PROB << " with workspace_min: " << Heuristics[abs_addr].x_min
			<< " workspace_max: " << Heuristics[abs_addr].x_max
			<< std::endl;
#endif
			count_rot++;
			break;
		case 6:
			//translation on x,y,z
#ifdef NDEBUG
			FS_PROB<<" on x_y_z direction";
#endif

			Samples_out[counter].xroll_S = Samples_in[relative_addr].xroll_S;
			Samples_out[counter].ypitch_S = Samples_in[relative_addr].ypitch_S;
			Samples_out[counter].zyaw_S = Samples_in[relative_addr].zyaw_S;
			diffuse_translation(Samples_in[relative_addr].x_S,
					Samples_out[counter].x_S, Heuristics[abs_addr].x_min,
					Heuristics[abs_addr].x_max, Trans_Rand);
			diffuse_translation(Samples_in[relative_addr].y_S,
					Samples_out[counter].y_S, Heuristics[abs_addr].y_min,
					Heuristics[abs_addr].y_max, Trans_Rand);
			diffuse_translation(Samples_in[relative_addr].z_S,
					Samples_out[counter].z_S, Heuristics[abs_addr].z_min,
					Heuristics[abs_addr].z_max, Trans_Rand);
#ifdef NDEBUG
			FS_PROB << " with workspace_min: " << Heuristics[abs_addr].x_min
			<< " workspace_max: " << Heuristics[abs_addr].x_max
			<< std::endl;
#endif
			count_trans += 3;
			break;
		case 7:
			//rotation on x y z
#ifdef NDEBUG
			FS_PROB <<" on x_y_z_rot direction";
#endif
			Samples_out[counter].x_S = Samples_in[relative_addr].x_S;
			Samples_out[counter].y_S = Samples_in[relative_addr].y_S;
			Samples_out[counter].z_S = Samples_in[relative_addr].z_S;

			diffuse_rotation(Samples_in[relative_addr].xroll_S,
					Samples_out[counter].xroll_S, Rot_Rand);
			diffuse_rotation(Samples_in[relative_addr].ypitch_S,
					Samples_out[counter].ypitch_S, Rot_Rand);
			diffuse_rotation(Samples_in[relative_addr].zyaw_S,
					Samples_out[counter].zyaw_S, Rot_Rand);
			count_rot += 3;
			break;
		case 8:
#ifdef NDEBUG
			FS_PROB <<" x_rot 90 ";
#endif
			Samples_out[counter].x_S = Samples_in[relative_addr].x_S;
			Samples_out[counter].y_S = Samples_in[relative_addr].y_S;
			Samples_out[counter].z_S = Samples_in[relative_addr].z_S;
			Samples_out[counter].ypitch_S = Samples_in[relative_addr].ypitch_S;
			Samples_out[counter].zyaw_S = Samples_in[relative_addr].zyaw_S;
			rot_type = uniform_single_generator(2);
			if (rot_type == 0) {
				TEMP_ADD = angle_Data_T(1.57);
			} else if (rot_type == 2) {
				TEMP_ADD = angle_Data_T(-1.57);
			} else if (rot_type == 1) {
				TEMP_ADD = angle_Data_T(3.14);
			} else if (rot_type == 3) {
				TEMP_ADD = angle_Data_T(-3.14);
			}
			temp_angle = ap_fixed<14 + 5, 6>(
					Samples_in[relative_addr].xroll_S + TEMP_ADD);
			round_angle(temp_angle);
#ifdef NDEBUG
			FS_PROB <<" old: "<< Samples_in[counter].xroll_S << " new: " << temp_angle <<" workspace_min: " << Heuristics[abs_addr].x_min <<" workspace_max: "<< Heuristics[abs_addr].x_max << std::endl;
#endif
			Samples_out[counter].xroll_S = temp_angle;
			break;
		case 9:
#ifdef NDEBUG
			FS_PROB << " y_rot 90 ";
#endif
			Samples_out[counter].x_S = Samples_in[relative_addr].x_S;
			Samples_out[counter].y_S = Samples_in[relative_addr].y_S;
			Samples_out[counter].z_S = Samples_in[relative_addr].z_S;
			Samples_out[counter].xroll_S = Samples_in[relative_addr].xroll_S;
			Samples_out[counter].zyaw_S = Samples_in[relative_addr].zyaw_S;
			rot_type = uniform_single_generator(2);
			if (rot_type == 0) {
				TEMP_ADD = angle_Data_T(1.57);
			} else if (rot_type == 2) {
				TEMP_ADD = angle_Data_T(-1.57);
			} else if (rot_type == 1) {
				TEMP_ADD = angle_Data_T(3.14);
			} else if (rot_type == 3) {
				TEMP_ADD = angle_Data_T(-3.14);
			}
			temp_angle = ap_fixed<14 + 5, 6>(
					Samples_in[relative_addr].ypitch_S + TEMP_ADD);
			round_angle(temp_angle);
#ifdef NDEBUG
			FS_PROB <<" old: "<< Samples_in[counter].ypitch_S << " new: " << temp_angle << " workspace_min: " << Heuristics[abs_addr].x_min << " workspace_max: "<< Heuristics[abs_addr].x_max <<std::endl;
#endif
			Samples_out[counter].ypitch_S = temp_angle;

			break;
		case 10:
#ifdef NDEBUG
			FS_PROB <<" z_rot 90 ";
#endif
			Samples_out[counter].x_S = Samples_in[relative_addr].x_S;
			Samples_out[counter].y_S = Samples_in[relative_addr].y_S;
			Samples_out[counter].z_S = Samples_in[relative_addr].z_S;
			Samples_out[counter].xroll_S = Samples_in[relative_addr].xroll_S;
			Samples_out[counter].ypitch_S = Samples_in[relative_addr].ypitch_S;

			rot_type = uniform_single_generator(2);
			if (rot_type == 0) {
				TEMP_ADD = angle_Data_T(1.57);
			} else if (rot_type == 2) {
				TEMP_ADD = angle_Data_T(-1.57);
			} else if (rot_type == 1) {
				TEMP_ADD = angle_Data_T(3.14);
			} else if (rot_type == 3) {
				TEMP_ADD = angle_Data_T(-3.14);
			}
			temp_angle = ap_fixed<14 + 5, 6>(
					Samples_in[relative_addr].zyaw_S + TEMP_ADD);
			round_angle(temp_angle);
#ifdef NDEBUG
			FS_PROB <<" old: "<< Samples_in[counter].zyaw_S << " new: " << temp_angle <<" workspace_min: " << Heuristics[abs_addr].x_min <<" workspace_max: "<< Heuristics[abs_addr].x_max <<std::endl;
#endif
			Samples_out[counter].zyaw_S = temp_angle;
			break;
		case 11:
			//Diffuse among current density map
			direction = uniform_single_generator(4);
#ifdef NDEBUG
			FS_PROB <<" diffuse current density map with direction: " << direction <<" abs_addr: " << abs_addr << " old h: "<< h_ind << " old w: " << w_ind;
#endif
			if (abs_addr >= HEATMAP_1 + HEATMAP_2) {
				offset = HEATMAP_1 + HEATMAP_2;
				map_width = HEATMAP_W3;
				map_height = HEATMAP_H3;
			} else if (abs_addr >= HEATMAP_1) {
				offset = HEATMAP_1;
				map_width = HEATMAP_W2;
				map_height = HEATMAP_H2;
			} else {
				map_width = HEATMAP_W1;
				map_height = HEATMAP_H1;
			}

			if (direction == 0) {
				temp_ind = ap_int<9>(h_ind - 1);
				if (temp_ind >= 0) {
					h_ind = temp_ind.range(7, 0);
				} else {
					h_ind = 0;
				}
			} else if (direction == 1) {
				temp_ind = h_ind + 1;
				if (temp_ind < map_height) {
					h_ind = temp_ind.range(7, 0);
				} else {
					h_ind = map_height - 1;
				}
			} else if (direction == 2) {
				temp_ind = w_ind - 1;
				if (temp_ind >= 0) {
					w_ind = temp_ind.range(7, 0);
				} else {
					w_ind = 0;
				}
			} else if (direction == 3) {
				temp_ind = w_ind + 1;
				if (temp_ind >= map_width) {
					w_ind = map_width - 1;
				} else {
					w_ind = temp_ind.range(7, 0);
				}
			}

			if (h_ind < 0) {
				h_ind = 0;
			} else if (h_ind >= map_height) {
				h_ind = map_height - 1;
			}
			if (w_ind < 0) {
				w_ind = 0;
			} else if (w_ind >= map_width) {
				w_ind = map_width - 1;
			}
			assert(h_ind >= 0);
			assert(h_ind < map_height);
			assert(w_ind >= 0);
			assert(w_ind < map_width);
#ifdef NDEBUG
			FS_PROB << "current h ind: " << h_ind << " current w ind: " << w_ind<<std::endl;
#endif
			new_abs_addr = offset + h_ind * map_width + w_ind;
			Sample_abs_indx_OUT[counter] = new_abs_addr;
			Samples_out[counter].x_S = Samples_in[relative_addr].x_S;
			Samples_out[counter].y_S = Samples_in[relative_addr].y_S;
			Samples_out[counter].z_S = Samples_in[relative_addr].z_S;
			Samples_out[counter].xroll_S = Samples_in[relative_addr].xroll_S;
			Samples_out[counter].ypitch_S = Samples_in[relative_addr].ypitch_S;
			Samples_out[counter].zyaw_S = Samples_in[relative_addr].zyaw_S;

			check_heuristic(Samples_out[counter], Heuristics[new_abs_addr]);
			break;
		case 12:
#ifdef NDEBUG
			FS_PROB <<" diffuse across heatmap with old abs_addr: " << abs_addr << " old h ind: "<< h_ind << " old w_ind: "<< w_ind;
#endif
			if (abs_addr >= HEATMAP_1 + HEATMAP_2) {
				offset = HEATMAP_1 + HEATMAP_2;
				//can only go to the smaller one
				w_ind = w_ind / 2;
				h_ind = h_ind / 2;
				new_abs_addr = HEATMAP_1 + h_ind * HEATMAP_W2 + w_ind;
			} else if (abs_addr >= HEATMAP_1) {
				offset = HEATMAP_1;
				map_width = HEATMAP_W2;
				map_height = HEATMAP_H2;
				direction = uniform_single_generator(2);
				if (direction == 0) {
					//go to the smaller heat map
					w_ind = w_ind / 2;
					h_ind = h_ind / 2;
					new_abs_addr = h_ind * HEATMAP_W1 + w_ind;

				} else {
					//go to the bigger heatmap
					w_ind = w_ind * 2;
					h_ind = h_ind * 2;
					new_abs_addr = (HEATMAP_1 + HEATMAP_2) + h_ind * HEATMAP_W3
							+ w_ind;
				}
			} else {
				//can only go to the bigger one
				w_ind = w_ind * 2;
				h_ind = h_ind * 2;
				new_abs_addr = HEATMAP_1 + h_ind * HEATMAP_W2 + w_ind;
			}
			Sample_abs_indx_OUT[counter] = new_abs_addr;
			Samples_out[counter].x_S = Samples_in[relative_addr].x_S;
			Samples_out[counter].y_S = Samples_in[relative_addr].y_S;
			Samples_out[counter].z_S = Samples_in[relative_addr].z_S;
			Samples_out[counter].xroll_S = Samples_in[relative_addr].xroll_S;
			Samples_out[counter].ypitch_S = Samples_in[relative_addr].ypitch_S;
			Samples_out[counter].zyaw_S = Samples_in[relative_addr].zyaw_S;

			check_heuristic(Samples_out[counter], Heuristics[new_abs_addr]);
#ifdef NDEBUG
			FS_PROB << " to new abs addr: " << new_abs_addr
			<< " with new workspace_min: "
			<< Heuristics[new_abs_addr].x_min
			<< " with new workspace_max: "
			<< Heuristics[new_abs_addr].x_max << std::endl;
#endif
			break;
		default:
#ifdef NDEBUG
			std::cout<<" WRONG!"<<std::endl;
#endif
			break;
		}
	}
	weight_address_Data_T REL_ADDR = Resampled_Addresses[509];
	Samples_out[509] = Samples_in[REL_ADDR];
	Sample_abs_indx_OUT[509] = Sample_abs_indx[REL_ADDR];

//deplete the trans and rot stream
	for (int i = 0; i < Gauss_number_needed - count_trans; i++) {
		Trans_Rand.read();
	}
	for (int i = 0; i < Gauss_number_needed - count_rot; i++) {
		Rot_Rand.read();
	}
#ifdef NDEBUG
	printf("EXIT_DIFFUSE\n");
	FS_PROB.close();
#endif

}

void diffuse_translation(xyz_Data_T t_in, xyz_Data_T &t_out,
		xyz_Data_T work_space_min, xyz_Data_T work_space_max,
		hls::stream<Gauss_Data_T> &Trans_Rand) {

	Gauss_Data_T delta = Trans_Rand.read();
	ap_fixed<14 + 5, 6> temp = ap_fixed<14 + 5, 6>(t_in + delta);

	if (temp < work_space_min) {
		t_out = work_space_min;
	} else if (temp > work_space_max) {
		t_out = work_space_max;
	} else {
		t_out = temp;
	}
#ifdef NDEBUG
	FS_PROB << " with rand value: " << delta << " old value: " << t_in
	<< " new value: " << t_out << " temp: " << temp
	<< " workspace min: " << work_space_min << " workspace max: "
	<< work_space_max << std::endl;
#endif

}

void round_angle(ap_fixed<14 + 5, 6> &temp) {
	if (temp > angle_Data_T(3.14)) {
		temp = ap_fixed<14 + 5, 6>(temp + angle_Data_T(-6.28));
	} else if (temp < angle_Data_T(-3.14)) {
		temp = ap_fixed<14 + 5, 6>(temp + angle_Data_T(6.28));
	}

}
void diffuse_rotation(angle_Data_T r_in, angle_Data_T &r_out,
		hls::stream<Gauss_Data_T> &Rot_Rand) {
	Gauss_Data_T delta = Rot_Rand.read();
	ap_fixed<14 + 5, 6> temp = ap_fixed<14 + 5, 6>(r_in + delta);
	round_angle(temp);

	r_out = temp;
#ifdef NDEBUG
	FS_PROB << " with rand value: " << delta << " old value: " << r_in
	<< " and new value: " << r_out << " temp: " << temp;
#endif
}

void check_heuristic(Mem_SAMPLE & sample, Mem_Heuristic & heuristic) {
	if (sample.x_S < heuristic.x_min) {
		sample.x_S = heuristic.x_min;
	} else if (sample.x_S > heuristic.x_max) {
		sample.x_S = heuristic.x_max;
	}
	if (sample.y_S < heuristic.y_min) {
		sample.y_S = heuristic.y_min;
	} else if (sample.y_S > heuristic.y_max) {
		sample.y_S = heuristic.y_max;
	}
	if (sample.z_S < heuristic.z_min) {
		sample.z_S = heuristic.z_min;
	} else if (sample.z_S > heuristic.z_max) {
		sample.z_S = heuristic.z_max;
	}
}

u2_Data_T uniform_single_generator(u3_Data_T val) {
	static u15_Data_T lfsr = 347;
	u1_Data_T b_15 = lfsr.get_bit(15 - 15);
	u1_Data_T b_14 = lfsr.get_bit(15 - 14);
	u1_Data_T new_bit = b_15 ^ b_14;
	lfsr = lfsr >> 1;
	lfsr.set_bit(14, new_bit);
	ap_ufixed<15, 0> rand_01;
	rand_01.range() = lfsr; // u 1 - > 0

	u2_Data_T fill = val * rand_01;

	return fill;
}
