#include "FT_Cre_Mod.hpp"
#ifdef NDEBUG

void MatrixMult_DEBUG(float A[M_DIM][M_DIM], float B[M_DIM][M_DIM], float C[M_DIM][M_DIM]) {
	// matrix multiplication of a A*B matrix
	for (int row = 0; row < M_DIM; row++) {
		for (int col = 0; col < M_DIM; col++) {
			float sum = 0;
			for (int scan = 0; scan < M_DIM; scan++) {
				sum += A[row][scan] * B[scan][col];
			}
			C[row][col] = sum;
		}
	}
}
#endif

void FT_Generator(Mem_SAMPLE Samples[NUM_PARTICLES],
		stream<MatFT_Data_T> & FT_stream) {
#pragma HLS INLINE
#pragma HLS DATAFLOW
	static stream<MatPC_Trunc_Data_T> PC_stream("PC_EXT_stream");
#pragma HLS STREAM variable=PC_stream depth=1 dim=1

	static stream<angle_Data_T> Angle_stream("Angle_stream");
#pragma HLS STREAM variable=Angle_stream depth=2 dim=1

	static stream<xyz_Data_T> Pose_stream("Pose_stream");
#pragma HLS STREAM variable=Pose_stream depth=16 dim=1

	static stream<MatC_Data_T> C_RightMat_stream("C_Static");
#pragma HLS STREAM variable=C_RightMat_stream depth =1 dim=1

	static stream<MatP_Data_T> P_LeftMat_stream("P_static");
#pragma HLS STREAM variable=P_LeftMat_stream depth =1 dim=1

	static stream<sincos_Struct> z_angle_stream("Zangle_CreateFT");
#pragma HLS STREAM variable=z_angle_stream depth =1 dim=1

	static stream<sincos_Struct> yx_angle_stream("YXangle_CreateFT");
#pragma HLS STREAM variable=yx_angle_stream depth =1 dim=1

	static stream<MatPC_Trunc_Data_T> PC_restream("PC_restream");
#pragma HLS STREAM variable=PC_restream depth =1 dim=1

	static stream<RowTrRz_Mat_Data_T> TrRz_stream("TrRz_CreateFT");
#pragma HLS STREAM variable=TrRz_stream depth = 1 dim=1

	static stream<Col_MatRyRx_Data_T> RyRx_stream("RyRx_stream");
#pragma HLS STREAM variable=RyRx_stream depth =1 dim=1

	static stream<Row_MatPCTrRz_Data_T> PCTrRz_stream("PCTrRz_stream");
#pragma HLS STREAM variable=PCTrRz_stream depth =1 dim=1

	FTGen_LD_Samples(Samples, Angle_stream, Pose_stream);
	PrjInit(P_LeftMat_stream);
	CamInit(C_RightMat_stream);
	PCNode(P_LeftMat_stream, C_RightMat_stream, PC_stream);
	AngleStrm(Angle_stream, z_angle_stream, yx_angle_stream);
	RestreamPC(PC_stream, PC_restream);
	TrRzInit(Pose_stream, TrRz_stream, z_angle_stream);
	PCTrRzNode(PC_restream, TrRz_stream, PCTrRz_stream);
	RyRxInit(yx_angle_stream, RyRx_stream);
	FTNode(PCTrRz_stream, RyRx_stream, FT_stream);

}

/***************************************************************************** PC INIT ZONE
 *
 *
 * THIS GENERATES THE PC MATRIX ROW BY ROW
 * EACH EACH ELEMENT IN THE ROW IS TRANSMITTED IN SERIES
 * BETWEEN TWO CONSECUTIVE ROWS THERE IS A LATECY
 *
 *
 */

//PC INIT ZONE 1.1
//
// THIS FUNCTION SET THE Projection MATRIX AND TRANSMIT IT TO THE MATRIX MULTIPLIER NODE
//
void PrjInit(stream<MatP_Data_T> & Left_stream) {
#pragma HLS INTERFACE ap_ctrl_none register port=return
	MatP_Data_T P[M_DIM][M_DIM];    // projection matrix
	Projection_MSet_Static(P);

	loop_flat: for (u5_Data_T ind_flt = 0; ind_flt < _2DIM; ind_flt++) {
#pragma HLS PIPELINE II=1
		// PUSH OUT THE P MATRIX ROW BY ROW  IN ORDER
		// 00 , 01 , 02 , 03 , 10 , 11 etc..
		Left_stream.write(P[ind_flt.range(3, 2)][ind_flt.range(1, 0)]);
	}
}
//PC INIT ZONE 1.2
void Projection_MSet_Static(MatP_Data_T P[M_DIM][M_DIM]) {
#pragma HLS INLINE
	// Initialize projection matrix
	P[0][0] = P_0_0;
	P[0][1] = P_0_1;
	P[0][2] = P_0_2;
	P[0][3] = P_0_3;

	P[1][0] = P_1_0;
	P[1][1] = P_1_1;
	P[1][2] = P_1_2;
	P[1][3] = P_1_3;

	P[2][0] = P_2_0;
	P[2][1] = P_2_1;
	P[2][2] = P_2_2;
	P[2][3] = P_2_3;

	P[3][0] = P_3_0	;
	P[3][1] = P_3_1	;
	P[3][2] = P_3_2	;
	P[3][3] = P_3_3	;

}

/****************************************************************************** LD SAMPLES ZONE
 *
 *
 * THIS STAGE LOADS THE SAMPLES FROM THE STAGE BEFORE // INIT OR RESAMPLED
 *
 *
 */

void FTGen_LD_Samples(Mem_SAMPLE Samples[NUM_PARTICLES],
		stream<angle_Data_T> & Angle_stream, stream<xyz_Data_T> & Pose_stream) {
#pragma HLS INTERFACE ap_ctrl_none register port=return
	u10_Data_T index_sampl = 1;
	u4_Data_T ind_sw = 0;

	Mem_SAMPLE Samples_Fill = Samples[0];
	Mem_SAMPLE Samples_Next_Fill;
	loop_flat: for (u13_Data_T flat_indx = 0; flat_indx < 6 * Num_Particles;
			flat_indx++) {
#pragma HLS PIPELINE II=1

		/*
		 * PUSH OUT THE INFORMATION WITH THIS ORDER
		 * ZANGLE - X
		 */
		switch (ind_sw) {
		case 0:

#ifdef NDEBUG
			std::cout << "SAMPLE : " << index_sampl << "\t" << ind_sw << ") ";
#endif
			Angle_stream << Samples_Fill.zyaw_S;
			Samples_Next_Fill = Samples[index_sampl];
			index_sampl++;
#ifdef NDEBUG
			std::cout << "angle z : " << Samples_Fill.zyaw_S << std::endl;
#endif
			break;
		case 1:
			//this is x
			Pose_stream << Samples_Fill.x_S;
#ifdef NDEBUG
			std::cout << "pose x: " << Samples_Fill.x_S << std::endl;
#endif
			break;
		case 2:
			Angle_stream << Samples_Fill.ypitch_S;
#ifdef NDEBUG
			std::cout << "angle y : " << Samples_Fill.ypitch_S << std::endl;
#endif
			break;
		case 3:
			// this is y
			Pose_stream << Samples_Fill.y_S;
#ifdef NDEBUG
			std::cout << "pose y " << Samples_Fill.y_S << std::endl;
#endif
			break;
		case 4:
			Angle_stream << Samples_Fill.xroll_S;
#ifdef NDEBUG
			out << "angle x: " << Samples_Fill.xroll_S << std::endl;
#endif
			break;
		case 5:
			// this is z
			Pose_stream << Samples_Fill.z_S;
			Samples_Fill = Samples_Next_Fill;
#ifdef NDEBUG
			std::cout << "pose z: " << Samples_Fill.x_S << std::endl;
#endif
			break;
		}

		if (ind_sw == 5) {
			ind_sw = 0;
#ifdef NDEBUG
			std::cout << "-----------------------------" << std::endl;
#endif
		} else {
			ind_sw++;
		}

	}
}

////////////////////////////////////LD SAMPLES ZONE  //////////////////////////

//PC INIT ZONE 2.1
//
// THIS FUNCTION SET THE CAMERA MATRIX AND TRANSMIT IT TO THE MATRIX MULTIPLIER NODE
//

void CamInit(stream<MatC_Data_T> & RightMat_stream) {
#pragma HLS INTERFACE ap_ctrl_none register port=return
	// Initialize camera view matrix
	MatC_Data_T C[M_DIM][M_DIM];    // camera view matrix

	Camera_MSet_Static(C);

	loop_flat: for (u7_Data_T ind_flt = 0; ind_flt < _3DIM; ind_flt++) {
#pragma HLS PIPELINE II=1
		// for each element of the left line  I need  the row
		// because I want a burst of 4 element  PC00 PC01 PC02 PC03 -
		// - wait - PCi00 PCi1 .. etc..
		RightMat_stream.write(C[ind_flt.range(3, 2)][ind_flt.range(1, 0)]);
	}
}

//PC INIT ZONE 2.2
void Camera_MSet_Static(MatC_Data_T C[M_DIM][M_DIM]) {
#pragma HLS INLINE
	C[0][0] = C_0_0;
	C[0][1] = C_0_1;
	C[0][2] = C_0_2;
	C[0][3] = C_0_3;

	C[1][0] = C_1_0;
	C[1][1] = C_1_1;
	C[1][2] = C_1_2;
	C[1][3] = C_1_3;

	C[2][0] = C_2_0;
	C[2][1] = C_2_1;
	C[2][2] = C_2_2;
	C[2][3] = C_2_3;

	C[3][0] = C_3_0;
	C[3][1] = C_3_1;
	C[3][2] = C_3_2;
	C[3][3] = C_3_3 ;
}

//PC INIT ZONE 3.1

void PCNode(stream<MatP_Data_T> & Left_stream,
		stream<MatC_Data_T> & Right_stream,
		stream<MatPC_Trunc_Data_T> & Out_stream) {
#pragma HLS INTERFACE ap_ctrl_none register port=return

	// IN THIS NODE THE RIGHT MATRIX IS READ 4 TIMES MORE THAN THE LEFT
	// FOR EACH POINT OF A LEFT ROW  WE READ THE ENTIRE RIGHT ROW
	// 00*00 + 01*10 + 02*20 + 03*30 so
	// bit size is (BIT_C_INT+BIT_C_FRAC-1)+(BIT_P_INT+BIT_P_FRAC-1) + 1 + 2 (the two additions)
	MatP_Data_T left_matrix;
	MatC_Data_T right_matrix;
	MatPC_Data_T sum_mult[M_DIM];
#pragma HLS ARRAY_PARTITION variable=sum_mult complete factor=1 dim=1
	MatPC_Data_T mult;
	loop_flat: for (u7_Data_T ind_flt = 0; ind_flt < _3DIM; ind_flt++) {
#pragma HLS PIPELINE II=1
		// I must store the partial sum of the multiplications
		// the overall push
		if (ind_flt.range(1, 0) == 0) { // like comparing to 0
			// read the P matrix so that we have x0 x1
			Left_stream.read(left_matrix);
		}
		// Sum
		Right_stream.read(right_matrix);
		mult = left_matrix * right_matrix;

		if (ind_flt.range(3, 2) == 0) { // like comparing to 0
			// Initialise
			// for each partial mult I must place it in the right place
			// O0 = P00 * C00
			// O1 = P00 * C01
			// O2 = P00 * C02
			// O3 = P00 * C03
			// O0 += P00 * C10
			// O1 += P00 * C11
			// O2 += P00 * C12
			// O3 += P00 * C13 ... until P00 * C3x
			sum_mult[ind_flt.range(1, 0)] = mult;
		} else if (ind_flt.range(3, 2) == M_DIM - 1) {// like compare it to M_DIM-1 = 3
		// Push result
			MatPC_Data_T PCOUT_FULL = sum_mult[ind_flt.range(1, 0)] + mult;
			MatPC_Trunc_Data_T PC_TRUNC = PCOUT_FULL;
			Out_stream.write(PC_TRUNC);
		} else {
			// Accumulate
			sum_mult[ind_flt.range(1, 0)] += mult;
		}
	}
}

////////////////////////////////////END PC INIT ZONE //////////////////////////

void AngleStrm(stream<angle_Data_T> & Angle_stream,
		stream<sincos_Struct> & z_angle_stream,
		stream<sincos_Struct> & yx_angle_stream) {
#pragma HLS INTERFACE ap_ctrl_none register port=return
	angle_Data_T angle_fill;
	sincos_Struct angles;
	u3_Data_T angle_ind = 0;
#ifdef NDEBUG
	std::cout << "angle_STREAM : " << std::endl;
#endif
	loop_flat: for (u11_Data_T flat_indx = 0; flat_indx < 3 * Num_Particles;
			flat_indx++) {
#pragma HLS PIPELINE II=1
		angle_fill = Angle_stream.read();
		angles.cosine = hls::cos(angle_fill);
		angles.sine = hls::sin(angle_fill);
#ifdef NDEBUG
		std::cout << "angle_fill : " << angle_fill << " cos : " << angles.cosine
		<< " sin : " << angles.sine << std::endl;
		std::cout << "angle_fill float : " << angle_fill.to_float()
		<< " cos float : " << std::cos(angle_fill.to_float())
		<< " sin float : " << std::sin(angle_fill.to_float())
		<< std::endl;
		if (std::abs(std::sin(angle_fill.to_float()) - angles.sine.to_float())
				>= 0.001) {
			std::cout << " Allert sine";
		}
		if (std::abs(std::cos(angle_fill.to_float()) - angles.cosine.to_float())
				>= 0.001) {
			std::cout << " Allert cosine";
		}
		std::cout << "-----------------------------" << std::endl;
#endif
		if (angle_ind == 0) {
			z_angle_stream.write(angles);
			angle_ind++;
		} else if (angle_ind == 1) {
			yx_angle_stream.write(angles);
			angle_ind++;
		} else {
			yx_angle_stream.write(angles);
			angle_ind = 0;
		}
	}
	//}
}

void RestreamPC(stream<MatPC_Trunc_Data_T> & PC_stream,
		stream<MatPC_Trunc_Data_T> & PC_restream) {
#pragma HLS INTERFACE ap_ctrl_none register port=return
	MatPC_Trunc_Data_T PC[M_DIM][M_DIM]; // Projection and camera viewpoint matrix

	flat_loop: for (u13_Data_T index_flat = 0; index_flat < NSAMPL_2DIM;
			index_flat++) {
#pragma HLS PIPELINE II=1
		if (index_flat < 16) {
			MatPC_Trunc_Data_T temp = PC_stream.read();
			PC[index_flat.range(3, 2)][index_flat(1, 0)] = temp;
			PC_restream.write(temp);
		} else {
			PC_restream.write(
					PC[index_flat.range(3, 2)][index_flat.range(1, 0)]);
		}
	}
#ifdef NDEBUG
	float C_fl[M_DIM][M_DIM]; // rigid-body and camera viewpoint matrix
	float P_fl[M_DIM][M_DIM];// rigid-body and camera viewpoint matrix

	C_fl[0][0] = C_0_0;
	C_fl[0][1] = C_0_1;
	C_fl[0][2] = C_0_2;
	C_fl[0][3] = C_0_3;

	C_fl[1][0] = C_1_0;
	C_fl[1][1] = C_1_1;
	C_fl[1][2] = C_1_2;
	C_fl[1][3] = C_1_3;

	C_fl[2][0] = C_2_0;
	C_fl[2][1] = C_2_1;
	C_fl[2][2] = C_2_2;
	C_fl[2][3] = C_2_3;

	C_fl[3][0] = C_3_0;
	C_fl[3][1] = C_3_1;
	C_fl[3][2] = C_3_2;
	C_fl[3][3] = C_3_3;

	// Initialize projection matrix
	P_fl[0][0] = P_0_0;
	P_fl[0][1] = P_0_1;
	P_fl[0][2] = P_0_2;
	P_fl[0][3] = P_0_3;

	P_fl[1][0] = P_1_0;
	P_fl[1][1] = P_1_1;
	P_fl[1][2] = P_1_2;
	P_fl[1][3] = P_1_3;

	P_fl[2][0] = P_2_0;
	P_fl[2][1] = P_2_1;
	P_fl[2][2] = P_2_2;
	P_fl[2][3] = P_2_3;

	P_fl[3][0] = P_3_0;
	P_fl[3][1] = P_3_1;
	P_fl[3][2] = P_3_2;
	P_fl[3][3] = P_3_3;

	float PC_fl[M_DIM][M_DIM];
	MatrixMult_DEBUG(P_fl, C_fl, PC_fl);

	std::cout << "PC Module" << std::endl;
	for (int row = 0; row < 4; row++) {
		std::cout << "| ";
		for (int col = 0; col < 4; col++) {
			std::cout << PC[row][col] << ", ";
		}

		std::cout << " |" << std::endl;
	}

	std::cout << "PC SW " << std::endl;
	for (int row = 0; row < 4; row++) {
		std::cout << "| ";
		for (int col = 0; col < 4; col++) {
			std::cout << PC_fl[row][col] << ", ";
		}

		std::cout << " |" << std::endl;
	}

	std::cout << "Error " << std::endl;

	for (int row = 0; row < 4; row++) {
		std::cout << "| ";
		for (int col = 0; col < 4; col++) {
			std::cout << (PC_fl[row][col] - PC[row][col].to_float()) << ", ";
			if (std::abs(PC_fl[row][col] - PC[row][col].to_float()) >= 0.1) {
				std::cout << "Allert";
			}
		}

		std::cout << " |" << std::endl;
	}

#endif
}

void TrRzInit(stream<xyz_Data_T> & Pose_stream,
		stream<RowTrRz_Mat_Data_T> & TrRz_stream,
		stream<sincos_Struct> & z_angle_stream) {
#pragma HLS INTERFACE ap_ctrl_none register port=return

	sincos_Struct z_angle_fill;
	trigono_Data_T cos_z_angle;
	trigono_Data_T sin_z_angle;
	RowTrRz_Mat_Data_T row_Temp[M_DIM];
#ifdef NDEBUG
	float X;
	float Y;
	float Z;
#endif
	flat_loop: for (u13_Data_T index_flat = 0; index_flat < NSAMPL_2DIM;
			index_flat++) {
#pragma HLS PIPELINE II=1
		xyz_Data_T pose_curr;
		RowTrRz_Mat_Data_T TEMP;
		// Initialize translation matrix
		switch (index_flat.range(3, 0)) {
		case 0:
			z_angle_fill = z_angle_stream.read();
			cos_z_angle = z_angle_fill.cosine;
			sin_z_angle = z_angle_fill.sine;
			pose_curr = Pose_stream.read(); // this is the x
			TEMP.row_0 = cos_z_angle;
			TEMP.row_1 = -sin_z_angle;
			TEMP.row_2 = (0);
			TEMP.row_3 = pose_curr;

			row_Temp[0].row_0 = TEMP.row_0;
			row_Temp[0].row_1 = TEMP.row_1;
			row_Temp[0].row_2 = TEMP.row_2;
			row_Temp[0].row_3 = TEMP.row_3;
			TrRz_stream.write(TEMP);
#ifdef NDEBUG
			X = pose_curr.to_float();
#endif

			break;
		case 1:
			pose_curr = Pose_stream.read(); // this is the y
			TEMP.row_0 = sin_z_angle;
			TEMP.row_1 = cos_z_angle;
			TEMP.row_2 = (0);
			TEMP.row_3 = pose_curr;
			row_Temp[1].row_0 = TEMP.row_0;
			row_Temp[1].row_1 = TEMP.row_1;
			row_Temp[1].row_2 = TEMP.row_2;
			row_Temp[1].row_3 = TEMP.row_3;
			TrRz_stream.write(TEMP);
#ifdef NDEBUG
			Y = pose_curr.to_float();
#endif
			break;
		case 2:
			pose_curr = Pose_stream.read(); // this is the z
			TEMP.row_0 = (0);
			TEMP.row_1 = (0);
			TEMP.row_2 = (1);
			TEMP.row_3 = pose_curr;

			row_Temp[2].row_0 = TEMP.row_0;
			row_Temp[2].row_1 = TEMP.row_1;
			row_Temp[2].row_2 = TEMP.row_2;
			row_Temp[2].row_3 = TEMP.row_3;
			TrRz_stream.write(TEMP);
#ifdef NDEBUG
			Z = pose_curr.to_float();
#endif
			break;
		case 3:
			TEMP.row_0 = 0;
			TEMP.row_1 = 0;
			TEMP.row_2 = 0;
			TEMP.row_3 = 1;
			row_Temp[3].row_0 = TEMP.row_0;
			row_Temp[3].row_1 = TEMP.row_1;
			row_Temp[3].row_2 = TEMP.row_2;
			row_Temp[3].row_3 = TEMP.row_3;
			TrRz_stream.write(TEMP);
#ifdef NDEBUG

			float Rz[M_DIM][M_DIM];   // z-axis rotation matrix
			float Tr[M_DIM][M_DIM];// translation matrix

			float TrRz[M_DIM][M_DIM];// translation matrix

			Rz[2][0] = 0;
			Rz[2][1] = 0;
			Rz[2][2] = 1;
			Rz[2][3] = 0;

			Rz[3][0] = 0;
			Rz[3][1] = 0;
			Rz[3][2] = 0;
			Rz[3][3] = 1;

			Rz[0][2] = 0;
			Rz[0][3] = 0;
			Rz[1][2] = 0;
			Rz[1][3] = 0;

			Rz[0][0] = cos_z_angle.to_float();
			Rz[0][1] = -sin_z_angle.to_float();

			Rz[1][0] = sin_z_angle.to_float();
			Rz[1][1] = cos_z_angle.to_float();

			for (int i = 0; i < M_DIM; i++) {
				for (int j = 0; j < M_DIM; j++) {
					if (i == j) {
						Tr[i][j] = 1;
					} else {
						Tr[i][j] = 0;
					}
				}
			}

			Tr[0][3] = X;
			Tr[1][3] = Y;
			Tr[2][3] = Z;

			MatrixMult_DEBUG(Tr, Rz, TrRz);

			std::cout << "TrRz Module" << std::endl;
			for (int row = 0; row < 4; row++) {
				std::cout << "| ";
				std::cout << row_Temp[row].row_0 << ", ";
				std::cout << row_Temp[row].row_1 << ", ";
				std::cout << row_Temp[row].row_2 << ", ";
				std::cout << row_Temp[row].row_3 << ", ";

				std::cout << " |" << std::endl;
			}
			std::cout << "TrRz SW" << std::endl;
			for (int row = 0; row < 4; row++) {
				std::cout << "| ";
				for (int col = 0; col < 4; col++) {
					std::cout << TrRz[row][col] << ", ";
				}

				std::cout << " |" << std::endl;
			}

			std::cout << "Error " << std::endl;

			for (int row = 0; row < 4; row++) {
				std::cout << "| ";

				std::cout << TrRz[row][0] - row_Temp[row].row_0.to_float()
				<< ", ";
				if (std::abs(
								TrRz[row][0] - row_Temp[row].row_0.to_float() >= 0.1)) {
					std::cout << "Allert";
				}
				std::cout << TrRz[row][1] - row_Temp[row].row_1.to_float()
				<< ", ";
				if (std::abs(
								TrRz[row][1] - row_Temp[row].row_1.to_float() >= 0.1)) {
					std::cout << "Allert";
				}
				std::cout << TrRz[row][2] - row_Temp[row].row_2.to_float()
				<< ", ";
				if (std::abs(
								TrRz[row][2] - row_Temp[row].row_2.to_float() >= 0.1)) {
					std::cout << "Allert";
				}
				std::cout << TrRz[row][3] - row_Temp[row].row_3.to_float()
				<< ", ";

				if (std::abs(
								TrRz[row][3] - row_Temp[row].row_3.to_float() >= 0.1)) {
					std::cout << "Allert";
				}

				std::cout << " |" << std::endl;

			}
			std::cout << std::endl;
#endif

			break;
		default:
			TrRz_stream.write(row_Temp[index_flat.range(1, 0)]);
			break;
		}

	}

}

void PCTrRzNode(stream<MatPC_Trunc_Data_T> & PC_stream,
		stream<RowTrRz_Mat_Data_T> & TrRz_stream,
		stream<Row_MatPCTrRz_Data_T> & PCTrRz_stream) {
#pragma HLS INTERFACE ap_ctrl_none register port=return
	RowTrRz_Mat_Data_T right_matrix;
	MatPC_Trunc_Data_T left_matrix;

	/*
	 struct RowTrRz_Mat_Data_T {
	 trigono_Data_T row_0; //ap_fixed<BIT_TRIG, BIT_TRIG_INT>  trigono_Data_T;
	 trigono_Data_T row_1;// ap_fixed<BIT_TRIG, BIT_TRIG_INT>  trigono_Data_T;
	 ap_ufixed<1,1> row_2; // 0nly 0 or 1
	 xyz_Data_T row_3; //ap_fixed<BIT_POSE, BIT_POSE_INT>  xyz_Data_T;// the bit sign

	 };


	 MatPC_Trunc_Data_T -> ap_ufixed<BIT_PC,BIT_PC_INT> ;

	 struct FR_Row_Mat_Data_T{
	 row_0 = ap_fixed<BIT_TRIG+BIT_PC, BIT_TRIG_INT+BIT_PC_INT-1+2>
	 row_1 = ap_fixed<BIT_TRIG+BIT_PC, BIT_TRIG_INT+BIT_PC_INT-1+2>
	 row_2 =  ap_ufixed<BIT_PC+2,BIT_PC_INT+2> ;
	 row_3 = ap_fixed<BIT_POSE++BIT_PC, BIT_POSE_INT+BIT_PC_INT-1+2>
	 }

	 WE MUST ADD 2 bit per four sum
	 WE must add the bit size for the multiplication
	 *
	 */
	FR_Row_Mat_Data_T sum_mult;

#ifdef NDEBUG
	float PC[M_DIM][M_DIM];
	float TrRz[M_DIM][M_DIM];
	FR_Row_Mat_Data_T PCTRz_Module[M_DIM];
#endif

	flat_loop: for (u13_Data_T index_flat = 0; index_flat < NSAMPL_2DIM;
			index_flat++) {
#pragma HLS PIPELINE II=1
#pragma HLS ALLOCATION instances=mul limit=4 operation

		PC_stream.read(left_matrix);
		TrRz_stream.read(right_matrix);

#ifdef NDEBUG
		PC[index_flat.range(3, 2)][index_flat.range(1, 0)] =
		left_matrix.to_float();
		TrRz[index_flat.range(1, 0)][0] = right_matrix.row_0.to_float();
		TrRz[index_flat.range(1, 0)][1] = right_matrix.row_1.to_float();
		TrRz[index_flat.range(1, 0)][2] = right_matrix.row_2.to_float();
		TrRz[index_flat.range(1, 0)][3] = right_matrix.row_3.to_float();

#endif

// Sum
		if (index_flat.range(1, 0) == 0) { // 0
			// Initialise
			sum_mult.row_0 = left_matrix * right_matrix.row_0;
			sum_mult.row_1 = left_matrix * right_matrix.row_1;
			sum_mult.row_2 = left_matrix * right_matrix.row_2;
			sum_mult.row_3 = left_matrix * right_matrix.row_3;

		} else if (index_flat.range(1, 0) == M_DIM - 1) { //3
// Store result
			Row_MatPCTrRz_Data_T trunc_res;
			sum_mult.row_0 += left_matrix * right_matrix.row_0;
			sum_mult.row_1 += left_matrix * right_matrix.row_1;
			sum_mult.row_2 += left_matrix * right_matrix.row_2;
			sum_mult.row_3 += left_matrix * right_matrix.row_3;
			trunc_res.row_0 = sum_mult.row_0;
			trunc_res.row_1 = sum_mult.row_1;
			trunc_res.row_2 = sum_mult.row_2;
			trunc_res.row_3 = sum_mult.row_3;
			PCTrRz_stream.write(trunc_res);
#ifdef NDEBUG
			PCTRz_Module[index_flat.range(3, 2)] = sum_mult;
			if (index_flat.range(3, 0) == 15) { // 15
				std::cout << "PC" << std::endl;
				for (int row = 0; row < 4; row++) {
					std::cout << "| ";
					for (int col = 0; col < 4; col++) {
						std::cout << PC[row][col] << ", ";
					}

					std::cout << " |" << std::endl;
				}
				std::cout << "TrRz" << std::endl;
				for (int row = 0; row < 4; row++) {
					std::cout << "| ";
					for (int col = 0; col < 4; col++) {
						std::cout << TrRz[row][col] << ", ";
					}

					std::cout << " |" << std::endl;
				}

				float PCTrRz[M_DIM][M_DIM];   // translation matrix

				MatrixMult_DEBUG(PC, TrRz, PCTrRz);

				std::cout << "PCTrRz Module" << std::endl;
				for (int row = 0; row < 4; row++) {
					std::cout << "| ";
					std::cout << PCTRz_Module[row].row_0 << ", ";
					std::cout << PCTRz_Module[row].row_1 << ", ";
					std::cout << PCTRz_Module[row].row_2 << ", ";
					std::cout << PCTRz_Module[row].row_3 << ", ";

					std::cout << " |" << std::endl;
				}
				std::cout << "PCTrRz SW" << std::endl;
				for (int row = 0; row < 4; row++) {
					std::cout << "| ";
					for (int col = 0; col < 4; col++) {
						std::cout << PCTrRz[row][col] << ", ";
					}

					std::cout << " |" << std::endl;
				}

				std::cout << "Error " << std::endl;

				for (int row = 0; row < 4; row++) {
					std::cout << "| ";

					std::cout
					<< PCTrRz[row][0]
					- PCTRz_Module[row].row_0.to_float()
					<< ", ";
					if (std::abs(
									PCTrRz[row][0] - PCTRz_Module[row].row_0.to_float()
									>= 0.1)) {
						std::cout << "Allert";
					}
					std::cout
					<< PCTrRz[row][1]
					- PCTRz_Module[row].row_1.to_float()
					<< ", ";
					if (std::abs(
									PCTrRz[row][1] - PCTRz_Module[row].row_1.to_float()
									>= 0.1)) {
						std::cout << "Allert";
					}
					std::cout
					<< PCTrRz[row][2]
					- PCTRz_Module[row].row_2.to_float()
					<< ", ";
					if (std::abs(
									PCTrRz[row][2] - PCTRz_Module[row].row_2.to_float()
									>= 0.1)) {
						std::cout << "Allert";
					}
					std::cout
					<< PCTrRz[row][3]
					- PCTRz_Module[row].row_3.to_float()
					<< ", ";
					if (std::abs(
									PCTrRz[row][3] - PCTRz_Module[row].row_3.to_float()
									>= 0.1)) {
						std::cout << "Allert";
					}

					std::cout << " |" << std::endl;
				}
			}
#endif

		} else {
			// Accumulate
			sum_mult.row_0 += left_matrix * right_matrix.row_0;
			sum_mult.row_1 += left_matrix * right_matrix.row_1;
			sum_mult.row_2 += left_matrix * right_matrix.row_2;
			sum_mult.row_3 += left_matrix * right_matrix.row_3;
		}
	}

}

void RyRxInit(stream<sincos_Struct> & yx_angle_stream,
		stream<Col_MatRyRx_Data_T> & RyRx_stream) {
#pragma HLS INTERFACE ap_ctrl_none register port=return

	FR_Col_MatRyRx_Data_T col_temp[M_DIM];
	sincos_Struct angle_fill;
	trigono_Data_T cos_y_angle;
	trigono_Data_T sin_y_angle;
	trigono_Data_T cos_x_angle;
	trigono_Data_T sin_x_angle;
	Col_MatRyRx_Data_T col_fill;
	flat_loop: for (u13_Data_T index_flat = 0; index_flat < NSAMPL_2DIM;
			index_flat++) {
#pragma HLS PIPELINE II=1
#pragma HLS ALLOCATION instances=mul limit=2 operation

		/*
		 * THE STRUCT
		 * row_0 cos*sin // two time trigo
		 * row_1 cos or sin // 1 time trigo
		 * row_2 cos*sin // 2 times bit trigo
		 * row_3 1 or 0 // 1 bit
		 */
// Initialize translation matrix
		switch (index_flat.range(3, 0)) {
		case 0:
			angle_fill = yx_angle_stream.read(); // THIS IS THE Y ANGLE
			cos_y_angle = angle_fill.cosine;
			sin_y_angle = angle_fill.sine;

			col_temp[0].col_0 = cos_y_angle;
			col_temp[0].col_1 = 0;
			col_temp[0].col_2 = -sin_y_angle;
			col_temp[0].col_3 = 0;
			col_fill.col_0 = col_temp[0].col_0;
			col_fill.col_1 = col_temp[0].col_1;
			col_fill.col_2 = col_temp[0].col_2;
			col_fill.col_3 = col_temp[0].col_3;
			break;
		case 1:
			angle_fill = yx_angle_stream.read(); // THIS IS THE X ANGLE

			cos_x_angle = angle_fill.cosine;
			sin_x_angle = angle_fill.sine;

			col_temp[1].col_0 = sin_y_angle * sin_x_angle;
			col_temp[1].col_1 = cos_x_angle;
			col_temp[1].col_2 = cos_y_angle * sin_x_angle;
			col_temp[1].col_3 = 0;
			col_fill.col_0 = col_temp[1].col_0;
			col_fill.col_1 = col_temp[1].col_1;
			col_fill.col_2 = col_temp[1].col_2;
			col_fill.col_3 = col_temp[1].col_3;
			break;
		case 2:
			col_temp[2].col_0 = sin_y_angle * cos_x_angle;
			col_temp[2].col_1 = -sin_x_angle;
			col_temp[2].col_2 = cos_y_angle * cos_x_angle;
			col_temp[2].col_3 = 0;

			col_fill.col_0 = col_temp[2].col_0;
			col_fill.col_1 = col_temp[2].col_1;
			col_fill.col_2 = col_temp[2].col_2;
			col_fill.col_3 = col_temp[2].col_3;

			break;
		case 3:
			col_temp[3].col_0 = 0;
			col_temp[3].col_1 = 0;
			col_temp[3].col_2 = 0;
			col_temp[3].col_3 = 1;

			col_fill.col_0 = col_temp[3].col_0;
			col_fill.col_1 = col_temp[3].col_1;
			col_fill.col_2 = col_temp[3].col_2;
			col_fill.col_3 = col_temp[3].col_3;

#ifdef NDEBUG

			float Rx[M_DIM][M_DIM];   // z-axis rotation matrix
			float Ry[M_DIM][M_DIM];// translation matrix

			Rx[0][0] = 1;
			Rx[0][1] = 0;
			Rx[0][2] = 0;
			Rx[0][3] = 0;

			Rx[3][0] = 0;
			Rx[3][1] = 0;
			Rx[3][2] = 0;
			Rx[3][3] = 1;

			Rx[1][0] = 0;
			Rx[1][3] = 0;
			Rx[2][0] = 0;
			Rx[2][3] = 0;

			Ry[3][0] = 0;
			Ry[3][1] = 0;
			Ry[3][2] = 0;
			Ry[3][3] = 1;

			Ry[1][0] = 0;
			Ry[1][1] = 1;
			Ry[1][2] = 0;
			Ry[1][3] = 0;

			Ry[0][1] = 0;
			Ry[0][3] = 0;
			Ry[2][1] = 0;
			Ry[2][3] = 0;

			Rx[1][1] = cos_x_angle.to_float();
			Rx[1][2] = -sin_x_angle.to_float();

			Rx[2][1] = sin_x_angle.to_float();
			Rx[2][2] = cos_x_angle.to_float();

			// Initialize y-axis rotation matrix
			Ry[0][0] = cos_y_angle.to_float();

			Ry[0][2] = sin_y_angle.to_float();

			Ry[2][0] = -sin_y_angle.to_float();

			Ry[2][2] = cos_y_angle.to_float();

			float RyRx[M_DIM][M_DIM];// translation matrix

			MatrixMult_DEBUG(Ry, Rx, RyRx);

			std::cout << "RyRx Module" << std::endl;

			std::cout << "| ";
			for (int row = 0; row < 4; row++) {
				std::cout << col_temp[row].col_0 << ", ";
			}
			std::cout << " |" << std::endl;
			std::cout << "| ";
			for (int row = 0; row < 4; row++) {
				std::cout << col_temp[row].col_1 << ", ";
			}
			std::cout << " |" << std::endl;
			std::cout << "| ";
			for (int row = 0; row < 4; row++) {
				std::cout << col_temp[row].col_2 << ", ";
			}
			std::cout << " |" << std::endl;
			std::cout << "| ";
			for (int row = 0; row < 4; row++) {
				std::cout << col_temp[row].col_3 << ", ";
			}
			std::cout << " |" << std::endl;
			std::cout << "RyRx SW" << std::endl;
			for (int row = 0; row < 4; row++) {
				std::cout << "| ";
				for (int col = 0; col < 4; col++) {
					std::cout << RyRx[row][col] << ", ";
				}

				std::cout << " |" << std::endl;
			}

			std::cout << "Error " << std::endl;

			std::cout << "| ";
			for (int col = 0; col < 4; col++) {
				std::cout << RyRx[0][col] - col_temp[col].col_0.to_float()
				<< ", ";
				if (std::abs(RyRx[0][col] - col_temp[col].col_0.to_float())
						>= 0.1) {
					std::cout << "Allert";
				}
			}
			std::cout << " |" << std::endl;
			std::cout << "| ";
			for (int col = 0; col < 4; col++) {
				std::cout << RyRx[1][col] - col_temp[col].col_1.to_float()
				<< ", ";
				if (std::abs(RyRx[1][col] - col_temp[col].col_1.to_float())
						>= 0.1) {
					std::cout << "Allert";
				}
			}
			std::cout << " |" << std::endl;
			std::cout << "| ";
			for (int col = 0; col < 4; col++) {
				std::cout << RyRx[2][col] - col_temp[col].col_2.to_float()
				<< ", ";
				if (std::abs(RyRx[2][col] - col_temp[col].col_2.to_float())
						>= 0.1) {
					std::cout << "Allert";
				}
			}
			std::cout << " |" << std::endl;
			std::cout << "| ";
			for (int col = 0; col < 4; col++) {
				std::cout << RyRx[3][col] - col_temp[col].col_3.to_float()
				<< ", ";
				if (std::abs(RyRx[3][col] - col_temp[col].col_3.to_float())
						>= 0.1) {
					std::cout << "Allert";
				}
			}

#endif
			break;
		default:

			col_fill.col_0 = col_temp[index_flat.range(1, 0)].col_0;
			col_fill.col_1 = col_temp[index_flat.range(1, 0)].col_1;
			col_fill.col_2 = col_temp[index_flat.range(1, 0)].col_2;
			col_fill.col_3 = col_temp[index_flat.range(1, 0)].col_3;
			break;
		}

		RyRx_stream.write(col_fill);

	}

}

void FTNode(stream<Row_MatPCTrRz_Data_T> & PCTrRz_stream,
		stream<Col_MatRyRx_Data_T> & RyRx_stream,
		stream<MatFT_Data_T> & FT_stream) {
#pragma HLS INTERFACE ap_ctrl_none register port=return
	MatFT_Data_T result;
	Col_MatRyRx_Data_T right_matrix;
	Row_MatPCTrRz_Data_T left_matrix;

	/*
	 *
	 struct Row_MatPCTrRz_Data_T {
	 ap_fixed<row_0_BIT_TR, BIT_TRIG_INT + BIT_PC_INT - 1 + 2> row_0;
	 ap_fixed<row_1_BIT_TR, BIT_TRIG_INT + BIT_PC_INT - 1 + 2> row_1;
	 ap_fixed<row_2_BIT_TR, BIT_PC_INT + 2> row_2;
	 ap_fixed<row_3_BIT_TR, BIT_POSE_INT + BIT_PC_INT - 1 + 2> row_3;
	 };
	 struct Col_MatRyRx_Data_T {
	 // all of these elements are lower or equal to 1
	 ap_fixed<BIT_TRIG + 5, BIT_TRIG_INT> col_0; // 2 times bit trigo
	 ap_fixed<BIT_TRIG, BIT_TRIG_INT> col_1; // 1 time trigo
	 ap_fixed<BIT_TRIG + 5, BIT_TRIG_INT> col_2; // 2 times bit trigo
	 ap_ufixed<1, 1> col_3; //
	 };

	 the col matrix is the rotation angle which is always <=1
	 result = row_0*col_0 + row1*col1 + ...
	 */
	flat_loop: for (u13_Data_T index_flat = 0; index_flat < NSAMPL_2DIM;
			index_flat++) {
#pragma HLS PIPELINE II=1
		FR_result result_part;
		if (index_flat.range(1, 0) == 0) {
			PCTrRz_stream.read(left_matrix);
		}

		RyRx_stream.read(right_matrix);

		result_part.result_0 = left_matrix.row_0 * right_matrix.col_0;
		result_part.result_1 = left_matrix.row_1 * right_matrix.col_1;
		result_part.result_2 = left_matrix.row_2 * right_matrix.col_2;
		result_part.result_3 = left_matrix.row_3 * right_matrix.col_3;
		result = result_part.result_0 + result_part.result_1
				+ result_part.result_2 + result_part.result_3;
		FT_stream.write(result);

#ifdef NDEBUG
		std::cout << 0 << ") " << left_matrix.row_0 << " * "
		<< right_matrix.col_0 << " = "
		<< left_matrix.row_0.to_float() * right_matrix.col_0.to_float()
		<< std::endl;
		std::cout << 1 << ") " << left_matrix.row_1 << " * "
		<< right_matrix.col_1 << " = "
		<< left_matrix.row_1.to_float() * right_matrix.col_0.to_float()
		<< std::endl;
		std::cout << 2 << ") " << left_matrix.row_2 << " * "
		<< right_matrix.col_2 << " = "
		<< left_matrix.row_2.to_float() * right_matrix.col_2.to_float()
		<< std::endl;
		std::cout << 3 << ") " << left_matrix.row_3 << " * "
		<< right_matrix.col_3 << " = "
		<< left_matrix.row_3.to_float() * right_matrix.col_3.to_float()
		<< std::endl;
		std::cout << "fp trunc" << result << std::endl;
		float result_flt = left_matrix.row_0.to_float()
		* right_matrix.col_0.to_float()
		+ left_matrix.row_1.to_float() * right_matrix.col_1.to_float()
		+ left_matrix.row_2.to_float() * right_matrix.col_2.to_float()
		+ left_matrix.row_3.to_float() * right_matrix.col_3.to_float();

		std::cout << "flt " << result_flt << std::endl;
		std::cout << "err " << std::abs(result_flt - result.to_float());
		if (std::abs(result_flt - result.to_float()) >= 0.1) {
			std::cout << "ALLERT " << std::endl;
		} else {
			std::cout << std::endl;

		}
#endif

	}

}
